import xml.etree.ElementTree as ET


def tagEditor(tag):
    return tag.replace("{http://www.gupro.de/GXL/gxl-1.0.dtd}","")

def checkErrorCode(faultNode, flagsName):
    for key in flagsName.keys():
        if key == faultNode:
            return flagsName.get(faultNode)
        else:
            return None

def preprocessing(learnerNodeName,flagsName, edgeName, faultName):
    # synchronising fualts and flags
    notErrorcodes = []
    for key in flagsName.keys():
        if key not in faultName:
            notErrorcodes.append(key)
    for item in notErrorcodes:
        del flagsName[item]

    # synchronising edges and faults
    notErrorEdges = []
    for item in edgeName:
        if item.get("to") not in flagsName.keys() or item.get("from")==learnerNodeName:
            notErrorEdges.append(item)
    for item in notErrorEdges:
        edgeName.remove(item)

def faultCodeTranslator(fault):
    if fault == "f001":
        return "A processing layer that operates on a N-dimensional tensors, should receive a valid input tensor with exactly N-dimensional shape(Missing reshape)."
    elif fault == "f002":
        return "Max-pooling is the preferred down-sampling strategy(lack of pooling)."
    elif fault == "f003":
        return "Multiple and redundant connected activations are not allowed."
    elif fault == "f004":
        return "The initialization of weights should not be constant to break the symmetry between neurons."
    elif fault == "f005":
        return "A learning layer should no longer include a bias when it is followed by batchnorm."
    elif fault == "f006":
        return "Batchnorm layer should be before the dropout."
    elif fault == "f007":
        return "The learnable parameters should be totally initialized once at the beginning of the training(multiple inits)"
    elif fault == "f008":
        return "A last layer activation is required to transform the logits into probabilities for classification problem(missing softmax)."
    elif fault == "f009":
        return "Activations for learning layers (i.e., convolution and fully-connected layer) should be a non-linear function."
    elif fault == "f010":
        return "The learnable parameters should be totally initialized once at the beginning of the training(missing init)"
    elif fault == "f011":
        return "A processing layer that operates on a N-dimensional tensors, should receive a valid input tensor with exactly N-dimensional shape(missing flatten )."
    elif fault == "f012":
        return "A last layer activation is required to transform the logits into probabilities for classification problem(bin missing sigmoid)."
    elif fault == "f013":
        return "A processing layer should receive sufficient-sized feature space to perform its spatial filtering or pooling."
    elif fault == "f014":
        return "mismatch shape : A reshape layer should never falsify the size of elements (i.e. first dimension)"
    elif fault == "f015":
        return "The loss should be correctly defined and connected to the layer in accordance with its input conditions (i.e.shape and type)-pre-activation"
    elif fault == "f016":
        return "The local window size for spatial filtering should generally increase or stay the same throughout the convolutional layers."
    elif fault == "f017":
        return "A last layer activation is required to transform the logits into probabilities for classification problem(bin wrong last layer activation)."
    elif fault == "f018":
        return "Dropout layer must be placed after the pooling layer to be more effective."
    elif fault == "f019":
        return "The area of feature maps and the width of fully-connected units should be progressively decreasing over the layers."
    elif fault == "f020":
        return "The loss should be correctly defined and connected to the layer in accordance with its input conditions (i.e.shape and type)-post_activation"
    elif fault == "f021":
        return "A reshape layer should preserve the total data elements/ A reshape layer should never falsify the size of elements (i.e. first dimension)."
    elif fault == "f022":
        return "The loss minimization problem should be solved iteratively with continuous update of parameters"
    elif fault == "f023":
        return "The number of feature maps should be gradually expanded while the feature map area is retracted."
    elif fault == "f024":
        return "The initialization of biases is preferred to be zeros."
    elif fault == "f025":
        return "Deep CNN should not apply pooling after every convolution."
    elif fault == "f026":
        return "Deep CNN should favor blocks of2,3or even 4 homogeneous convolutional layers with similar characteristics."
    elif fault == "f027":
        return "Max-pooling is the preferred down-sampling strategy(ineffective pooling)"
        
    else:
        return None


def main(inputGraphName="TF_lenet", parserType = "tf"):
    learnerNode = ""
    architectureNode = ""
    flags = {}
    edges = []
    faults = []
    layers = []
    nodeName = {}

    grooveOutputName = inputGraphName
    if parserType == "tf":
        graphFolderName = "TF_graphs"
    elif parserType == "keras":
        graphFolderName = "Keras_graphs"
    tree = ET.parse(f'{graphFolderName}/{grooveOutputName}.gst')
    root = tree.getroot()


    for edge in root.iter("{http://www.gupro.de/GXL/gxl-1.0.dtd}edge"):
        fromNode = edge.attrib.get("from")
        toNode = edge.attrib.get("to")
        flag = False
        if fromNode == toNode:
            for strTag in edge.iter("{http://www.gupro.de/GXL/gxl-1.0.dtd}string"):
                if strTag.text == "type:Learner":
                    learnerNode = fromNode
                elif strTag.text == "type:Architecture":
                    architectureNode = fromNode
                elif strTag.text == "type:Layer":
                    layers.append(fromNode)
                elif strTag.text.find("let:No = int:") != -1:
                    nodeName.update({fromNode: strTag.text.replace("let:No = int:", "")})
                elif strTag.text == "type:Faults":
                    faults.append(fromNode)
                elif strTag.text.find("flag:") != -1:
                    if fromNode in flags.keys():
                        temp = flags.get(fromNode)
                        temp.append(strTag.text.replace("flag:", ""))
                        flags.update({fromNode: temp})
                    else:
                        flags.update({fromNode: [strTag.text.replace("flag:", "")]})

        elif fromNode != toNode:
            for strTag in edge.iter("{http://www.gupro.de/GXL/gxl-1.0.dtd}string"):
                if strTag.text == "has-a" or strTag.text == "has":
                    directiveEdge = {"from": fromNode, "to": toNode}
                    edges.append(directiveEdge)


    # analyzing the result to relate faults and nodes
    buggyFlag = False
    output = ""

    # analyzing learner node
    learnerFaults = []
    architectureFault = []
    layerErrors = []
    layersFaults = {}
    for edge in edges:
        if edge.get("from") == learnerNode and edge.get("to") in faults and edge.get("to") in flags.keys():
            for item in flags.get(edge.get("to")):
                learnerFaults.append(item)
            edges.pop(edges.index(edge))
        if edge.get("from") == architectureNode and edge.get("to") in faults and edge.get("to") in flags.keys():
            for item in flags.get(edge.get("to")):
                architectureFault.append(item)
            edges.pop(edges.index(edge))


    if len(learnerFaults) != 0:
        buggyFlag = True
        output += "Learner ==> "

        for item in learnerFaults:
            output += faultCodeTranslator(item)
            if learnerFaults.index(item) != len(learnerFaults)-1:
                output += ", "

    if len(architectureFault) != 0:
        buggyFlag = True
        output += "Architecture ==> "
        for item in architectureFault:
            output += faultCodeTranslator(item)
            if architectureFault.index(item) != len(architectureFault) - 1:
                output += ", "

    preprocessing(learnerNode,flags, edges, faults)
    if buggyFlag and len(edges) != 0:
        output += "\n\r"

    if len(edges) != 0:
        buggyFlag = True

    buggyLayers = {}

    for item in edges:
        BuggyLayerName = item.get("from")
        buggyLayerNumber = nodeName.get(BuggyLayerName)
        layerErrors = flags.get(item.get("to"))
        report = "Layer {} ==> ".format(buggyLayerNumber)
        for item in layerErrors:
            report += faultCodeTranslator(item)
            if layerErrors.index(item) != len(layerErrors)-1:
                report += ", "

        buggyLayers.update({buggyLayerNumber:report})

    if len(buggyLayers) != 0:
        sortedBuggyLayers = dict(sorted(buggyLayers.items(), key=lambda s: int(s[0])))
        lastKey = list(sortedBuggyLayers.keys())[-1]
        for item in sortedBuggyLayers:
            output += sortedBuggyLayers.get(item)
            if item != lastKey:
                output += "\n\r"

    if not buggyFlag:
        output = "There is no identified fault in the DNN script"

    return output

if __name__ == "__main__":
    program_name = 'SO_deep_CNN_asymetry_blocksGrooveOut'
    result = main(program_name, "keras")
    print("result is :")
    print(result)



import numpy as np
import os

def fill_in_params(params, args_names):
    new_params = {}
    for key in params.keys():
        if isinstance(key, int):
            new_params[args_names[key]] = params[key]
        else:
            new_params[key] = params[key]
    return new_params

def is_DNN_layer(layer_type):
    supported_layer_types = ['Conv1D', 'Conv2D', 'Conv3D',
                             'MaxPooling1D', 'MaxPooling2D', 'MaxPooling3D',
                             'AveragePooling1D', 'AveragePooling2D', 'AveragePooling3D', 
                             'BatchNormalization', 'Flatten', 'Dense', 'Dropout', 'Activation'] 
                             
    lower_layer_type = layer_type.lower()
    if layer_type in supported_layer_types or lower_layer_type in Activation.activations:
        return True
    return False

def is_loss(function):
    losses = Loss.regression_losses + Loss.classification_losses
    if function in losses:
        return True
    return False

def is_optimizer(function):
    optimizers = list(Optimizer.optimizers.values())
    if function in optimizers:
        return True
    return False

class DNN_Program:

    def __init__(self, DNN, learner):
        self.DNN = DNN
        self.learner = learner

    def set_DNN(self, DNN):
        self.DNN = DNN
    
    def set_Learner(self, learner):
        self.learner = learner

    def gen_graph_nodes_and_edges(self):
        self.graph_nodes = ""
        self.graph_edges = ""
        node_counter = 0
        node_id = "n" + str(node_counter)
        self.graph_nodes += "\t\t<node id=\"{}\"/>\n".format(node_id)
        # DNN_program edges
        self.graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:DNN-program</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id)
        node_counter += 1
        previous_node = node_id
        node_id = "n" + str(node_counter)
        self.graph_nodes += "\t\t<node id=\"{}\"/>\n".format(node_id)
        # Data Edge
        self.graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            previous_node, node_id)
        self.graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Data</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id)
        node_counter += 1
        node_id = "n" + str(node_counter)
        self.graph_nodes += "\t\t<node id=\"{}\"/>\n".format(node_id)
        self.graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
        previous_node, node_id)
        node_counter, added_graph_nodes, added_graph_edges = self.DNN.gen_graph_nodes_and_edges(node_id, node_counter)
        self.graph_nodes += added_graph_nodes
        self.graph_edges += added_graph_edges
        node_counter += 1
        node_id = "n" + str(node_counter)
        self.graph_nodes += "\t\t<node id=\"{}\"/>\n".format(node_id)
        self.graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                            previous_node, node_id)
        node_counter, added_graph_nodes, added_graph_edges = self.learner.gen_graph_nodes_and_edges(node_id, node_counter)
        self.graph_nodes += added_graph_nodes
        self.graph_edges += added_graph_edges

    def write_graph_node(self, filename):
        filename = filename.split(os.path.sep)[-1]
        with open(f'Keras_graphs/{filename}.gst', 'w') as f:
            f.write("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n<gxl xmlns=\"http://www.gupro.de/GXL/gxl-1.0.dtd\">\n")
            f.write("\t<graph role=\"graph\" edgeids=\"false\" edgemode=\"directed\" id=\"start1\">\n")
            f.write(self.graph_nodes)
            f.write(self.graph_edges)
            # Finalizing the gst file
            f.write("\t</graph>\n</gxl>")

class Learner:

    def __init__(self):
        self.metric = 'mean_absolute_error'
        self.batch_size = 32
        self.epochs = 100

    def set_vars_init_location(self, vars_init_location):
        self.vars_init_location = vars_init_location

    def set_train_loop(self, train_loop):
        self.train_loop = train_loop

    def set_loss(self, loss):
        self.loss = loss
    
    def set_optimizer(self, optimizer):
        self.optimizer = optimizer
    
    def set_metric(self, metric):
        self.metric = metric
    
    def set_batch_size(self, batch_size):
        self.batch_size = batch_size
    
    def set_epochs(self, epochs):
        self.epochs = epochs
    
    def gen_graph_nodes_and_edges(self, node_id, node_counter):
        graph_nodes = ""
        graph_edges = ""
        #Main
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Learner</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id)
        if hasattr(self, 'train_loop'):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:loop_line_start = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                    node_id, node_id, self.train_loop.start_loc)
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:loop_line_end = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                    node_id, node_id, self.train_loop.end_loc)
        if hasattr(self, 'vars_init_location'):
            is_there_global_init = True if self.vars_init_location!=None else False
            if is_there_global_init:
                graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:global_initializer = bool:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                    node_id, node_id, str(is_there_global_init).lower())
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:location_of_initializer = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, self.vars_init_location)
        #Loss
        node_counter += 1
        previous_node = node_id
        node_id = "n" + str(node_counter)
        graph_nodes += "\t\t<node id=\"{}\"/>\n".format(node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            previous_node, node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Loss</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:type = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, self.loss.name)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:predictions = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, self.loss.predictions_layer)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:input_type = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, self.loss.input_type)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:problem_type = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, self.loss.problem_type)
        #Optimizer
        node_counter += 1
        node_id = "n" + str(node_counter)
        graph_nodes += "\t\t<node id=\"{}\"/>\n".format(node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            previous_node, node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Optimizer</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:type = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, self.optimizer.name)
        #graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:learning_rate = real:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
        #        node_id, node_id, self.optimizer.learning_rate)
        #graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:initializer = bool:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
        #        node_id, node_id, self.optimizer.initializer)
        #Hyperaparameters
        node_counter += 1
        node_id = "n" + str(node_counter)
        graph_nodes += "\t\t<node id=\"{}\"/>\n".format(node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            previous_node, node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Hyperparameters</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:batchSize = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, self.batch_size)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:epochCount = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, self.epochs)
        #Metric
        node_counter += 1
        node_id = "n" + str(node_counter)
        graph_nodes += "\t\t<node id=\"{}\"/>\n".format(node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            previous_node, node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Metric</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:type = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, self.metric)
        return node_counter, graph_nodes, graph_edges
 
class Train_Loop:

    def __init__(self, start_loc, end_loc):
        self.start_loc = start_loc
        self.end_loc = end_loc

class Optimizer:
    optimizers = {'sgd':'SGD', 'rmsprop':'RMSprop', 
                  'adagrad':'Adagrad', 'adam':'Adam', 'adadelta':'Adadelta'}
    params = ['learning_rate']

    def __init__(self, name, learning_rate=None, additional_params=None):
        self.name = self.clean(name)
        self.learning_rate = learning_rate
        self.additional_params = additional_params
    
    def clean(self, name):
        for short_name, long_name in Optimizer.optimizers.items():
            if name == long_name:
                return short_name
 
class Loss:

    regression_losses = ['hinge', 'huber_loss', 'mean_squared_error', 'mean_absolute_error']
    classification_losses = ['binary_crossentropy', 'categorical_crossentropy', 'sparse_categorical_crossentropy']
    params = {loss:['labels', 'predictions'] for loss in regression_losses}
    params.update({loss:['labels', 'logits'] if loss !='log_loss' else ['labels', 'predictions'] for loss in classification_losses})

    def __init__(self, name, predictions):
        self.name = name
        self.predictions = predictions
        self.infer_problem_type()
        self.infer_input_type()
        
    def set_predictions_layer(self, predictions_layer):
        self.predictions_layer = predictions_layer

    def infer_input_type(self):
        if self.problem_type == 'regression':
            self.input_type = 'pre_activation'
        elif self.problem_type == 'classification':
            self.input_type = 'post_activation'
            #if Loss.params[self.name][1] == 'predictions':
                #self.input_type = 'post_activation'
            #elif Loss.params[self.name][1] == 'logits':
                #self.input_type = 'pre_activation'
    
    def infer_problem_type(self):
        if self.name in Loss.regression_losses:
            self.problem_type = 'regression'
        elif self.name in Loss.classification_losses:
            self.problem_type = 'classification'

class Metric:

    def __init__(self, name):
        self.name = name
        self.infer_problem_type()
        self.infer_input_type()

    def infer_problem_type(self):
        pass
    
    def infer_input_type(self):
        pass

class DNN:

    def __init__(self, input_size, output_size):
        self.input_size = input_size
        self.output_size = output_size
        self.layers = []
        self.convs_count = 0
        self.pools_count = 0 
    
    def append_layer(self, layer_name, layer_type, params):
        if layer_type in ['Conv1D', 'Conv2D', 'Conv3D']:
            params_dict = fill_in_params(params, args_names=Convolutional.params)
            conv = Convolutional(layer_name, params_dict['inputs'], params_dict) 
            self.convs_count += 1
            self.layers.append(conv)  
        elif layer_type in ['MaxPooling1D', 'MaxPooling2D', 'MaxPooling3D',
                            'AveragePooling1D', 'AveragePooling2D', 'AveragePooling3D',
                            'MaxPool1D', 'MaxPool2D', 'MaxPool3D']:
            params_dict = fill_in_params(params, args_names=Pooling.params)
            pool = Pooling(layer_name, params_dict['inputs'], params_dict)
            self.pools_count += 1
            self.layers.append(pool)  
        elif layer_type == 'Flatten':
            params_dict = fill_in_params(params, args_names=Flatten.params)
            flat = Flatten(layer_name, params_dict['inputs'])
            self.layers.append(flat)  
        elif layer_type == 'Reshape':
            params_dict = fill_in_params(params, args_names=Reshape.params)
            reshape = Reshape(layer_name, params_dict['tensor'], params_dict['shape'])
            self.layers.append(reshape)  
        elif layer_type == 'Dense':
            params_dict = fill_in_params(params, args_names=Dense.params)
            dense = Dense(layer_name, params_dict['inputs'], params_dict)
            self.layers.append(dense)  
        elif layer_type == 'Dropout':
            params_dict = fill_in_params(params, args_names=Dropout.params)
            dropout = Dropout(layer_name, params_dict['inputs'], params_dict['rate'])
            self.layers.append(dropout)  
        elif layer_type == 'BatchNormalization':
            params_dict = fill_in_params(params, args_names=Batchnorm.params)
            batchnorm = Batchnorm(layer_name, params_dict['inputs'])
            self.layers.append(batchnorm)  
        elif layer_type == 'Activation':
            layer_type = params[0]
            inputs = params['inputs']
            activation = Activation(layer_name, inputs, layer_type)
            self.layers.append(activation)
        elif layer_type.lower() in Activation.activations:
            params_dict = fill_in_params(params, args_names=Activation.params)
            inputs = params_dict['inputs']
            activation = Activation(layer_name, inputs, layer_type)
            self.layers.append(activation)  

    def update_layers_name(self, old_name, new_name):
        for layer in self.layers:
            if layer.name == old_name:
                layer.name = new_name

    def compile(self):
        def _layer_idx(layer_name, connected_layers):
            for layer_idx in range(len(connected_layers)):
                if connected_layers[layer_idx].name == layer_name:
                    return layer_idx
        connected_layers = []
        input_layer_name = self.layers[0].input_layer_name
        input_size = self.input_size
        inp_layer_names = [layer.input_layer_name for layer in self.layers]
        reversed_inp_layer_names = list(reversed(inp_layer_names))
        idx = 0
        while idx != len(reversed_inp_layer_names) - 1:
            idx = (len(reversed_inp_layer_names)-1) - reversed_inp_layer_names.index(input_layer_name)
            self.layers[idx].set_input_size(input_size)
            input_size = self.layers[idx].output_size
            input_layer_name = self.layers[idx].name
            connected_layers.append(self.layers[idx])
        layers_to_add = {}
        for layer_idx in range(len(connected_layers)):
            layer = connected_layers[layer_idx]
            if isinstance(layer, Convolutional) or isinstance(layer, Dense):
                if layer.activation:
                    activation_layer = Activation(f'act_{layer.name}', layer.name, layer.activation)
                    activation_layer.set_input_size(layer.output_size)
                    layers_to_add[layer.name] = activation_layer       
        for layer_name, layer_to_add in layers_to_add.items():
            layer_idx = _layer_idx(layer_name, connected_layers)
            connected_layers.insert(layer_idx+1, layer_to_add)
        self.layers = connected_layers

    def get_last_layer_idx(self):
        return len(self.layers)

    def gen_graph_nodes_and_edges(self, node_id, node_counter):
        graph_nodes = ""
        graph_edges = ""
        #Main
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Architecture</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:hiddenLayerCount = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.layers))
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:ConvCount = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, self.convs_count)  
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:PoolCount = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, self.pools_count)
        #Input Layer
        node_counter += 1
        previous_node = node_id
        node_id = "n" + str(node_counter)
        graph_nodes += "\t\t<node id=\"{}\"/>\n".format(node_id)
        layer_counter = 0
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>starts</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            previous_node, node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:InputLayer</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:No = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, layer_counter)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.input_size))
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:dim1 = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, self.input_size[1])
        if len(self.input_size) == 3:
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:channels = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, self.input_size[2])
        elif len(self.input_size) == 4:
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:dim2 = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, self.input_size[2])
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:channels = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, self.input_size[3])
        elif len(self.input_size) == 5:
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:dim2 = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, self.input_size[2])
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:dim3 = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, self.input_size[3])
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:channels = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, self.input_size[4])
        # Hidden Layers
        for layer_idx in range(len(self.layers)):
            layer = self.layers[layer_idx]
            layer_counter += 1
            node_counter += 1
            previous_node = node_id
            node_id = "n" + str(node_counter)
            graph_nodes += "\t\t<node id=\"{}\"/>\n".format(node_id)
            if layer_idx == 0:
                graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>followed-by</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                        previous_node, node_id)
            else:
                graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>next</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                    previous_node, node_id)
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:No = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, layer_counter)
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Layer</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id)
            if layer_idx == len(self.layers) - 1:
                graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>flag:output</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                    node_id, node_id)
            node_counter, added_graph_nodes, added_graph_edges = layer.gen_graph_nodes_and_edges(node_id, node_counter)
            graph_nodes += added_graph_nodes
            graph_edges += added_graph_edges 
        #Labels Layer
        node_counter += 1
        previous_node = node_id
        node_id = "n" + str(node_counter)
        graph_nodes += "\t\t<node id=\"{}\"/>\n".format(node_id)
        layer_counter = 0
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>ends</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            previous_node, node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Labels</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:size = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, self.output_size[1])
        return node_counter, graph_nodes, graph_edges
    
    def __str__(self):
        s = ''
        for layer in self.layers:
            s += f'{layer.input_layer_name}== {layer.__class__.__name__} ==> {layer.name}'
            if hasattr(layer, "output_size"):
                s += f'|({layer.input_size} ==> {layer.output_size})'
            s += ' \n'
        return s

class Processing_Layer:
    
    def __init__(self, name, input_layer_name):
        self.name = name
        self.input_layer_name = input_layer_name

class Learning_Layer(Processing_Layer):
    
    def __init__(self, name, input_layer_name):
        super(Learning_Layer, self).__init__(name, input_layer_name)
        
class Convolutional(Learning_Layer):
    params = ['filters', 'kernel_size', 'strides', 'padding', 'inputs']

    def __init__(self, name, input_layer_name, params):
        super(Convolutional, self).__init__(name, input_layer_name)
        self.filters = params['filters'] if 'filters' in params else None
        self.set_kernel_size(params)
        self.set_strides(params)
        self.padding = params['padding'] if 'padding' in params else 'valid'
        self.activation = params['activation'] if 'activation' in params else None
        self.use_bias = params['use_bias'] if 'use_bias' in params else True
        self.W_init = params['kernel_initializer'] if 'kernel_initializer' in params else 'glorot_uniform'
        self.b_init = params['bias_initializer'] if 'bias_initializer' in params else 'zeros'
        self.W_cons = params['kernel_constraint'] if 'kernel_constraint' in params else None
        self.b_cons = params['bias_constraint'] if 'bias_constraint' in params else None
        self.W_reg = params['kernel_regularizer'] if 'kernel_regularizer' in params else None
        self.b_reg = params['bias_regularizer'] if 'bias_regularizer' in params else None
        self.clean_strings_names()
    
    def clean_strings_names(self):
        if self.activation:
            self.activation = self.activation.split('.')[-1]
        if self.W_init:
            self.W_init = self.W_init.split('.')[-1]
            if self.W_init in ['zeros', 'ones']:
                self.W_init = 'constant'
        if self.b_init:
            self.b_init = self.b_init.split('.')[-1]
        if self.W_cons:
            self.W_cons = self.W_cons.split('.')[-1]
        if self.b_cons:
            self.b_cons = self.b_cons.split('.')[-1]
        if self.W_reg:
            self.W_reg = self.W_reg.split('.')[-1]
        if self.b_reg:
            self.b_reg = self.b_reg.split('.')[-1]

    def set_kernel_size(self, params):
        kernel_size = params['kernel_size'] if 'kernel_size' in params else None
        if isinstance(kernel_size, list):
            self.kernel_size = kernel_size
        elif isinstance(kernel_size, tuple):
            self.kernel_size = [kernel_size[0], kernel_size[1]]    
        else:
            self.kernel_size = [kernel_size, kernel_size]

    def set_strides(self, params):
        strides = params['strides'] if 'strides' in params else (1, 1)
        if isinstance(strides, list):
            self.strides = strides
        elif isinstance(strides, tuple):
            self.strides = [strides[0], strides[1]]
        else:
            self.strides = [strides, strides]

    def set_input_size(self, input_size):
        self.input_size = input_size
        self.output_size = self.infer_output_size()

    def infer_output_size(self):
        if len(self.input_size) == 4:
            batch_size, in_height, in_width, in_depth = self.input_size
            if self.padding == 'valid':
                out_height = (in_height - self.kernel_size[0]) / self.strides[0] + 1
                out_width = (in_width - self.kernel_size[1]) / self.strides[1] + 1
            elif self.padding == 'same':
                out_height = (in_height - 1) / self.strides[0] + 1
                out_width = (in_width - 1) / self.strides[1] + 1
            elif self.padding == 'full':
                out_height = (in_height + self.kernel_size[0] - 2) / self.strides[0] + 1
                out_width = (in_width + self.kernel_size[1] - 2) / self.strides[1] + 1
            return [batch_size, int(out_height), int(out_width), self.filters]
        elif len(self.input_size) == 3:
            batch_size, in_steps, in_depth = self.input_size
            if self.padding == 'valid':
                out_steps = (in_steps - self.kernel_size[0]) / self.strides[0] + 1
            elif self.padding == 'same':
                out_steps = (in_steps - 1) / self.strides[0] + 1
            elif self.padding == 'full':
                out_steps = (in_steps + self.kernel_size[0] - 2) / self.strides[0] + 1
            return [batch_size, int(out_steps), self.filters]
    
    def gen_graph_nodes_and_edges(self, node_id, node_counter):
        graph_nodes = ""
        graph_edges = ""
        #Layer
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:in_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.input_size))
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.output_size))
        for i in range(len(self.input_size)):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:in_dim{} = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, i, self.input_size[i])
        for i in range(len(self.output_size)):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dim{} = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, i, self.output_size[i])
        node_counter += 1
        next_node_id = "n" + str(node_counter)
        graph_nodes += "\t\t<node id=\"{}\"/>\n".format(next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has-a</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Parameters</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:type = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.name.split('_')[0])
        
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:filters = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.filters)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:strides_dim1 = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.strides[0])
        if 'conv2d' in self.name or 'conv3d' in self.name:
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:strides_dim2 = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                    next_node_id, next_node_id, self.strides[1])
        if 'conv3d' in self.name:
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:strides_dim3 = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                    next_node_id, next_node_id, self.strides[2])
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:kernel_dim1 = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.kernel_size[0])
        if 'conv2d' in self.name or 'conv3d' in self.name:
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:kernel_dim2 = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                    next_node_id, next_node_id, self.kernel_size[1])
        if 'conv3d' in self.name:
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:kernel_dim3 = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                    next_node_id, next_node_id, self.kernel_size[2])
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:padding = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.padding)
        #additional Attributes
        node_counter += 1
        next_node_id = "n" + str(node_counter)
        graph_nodes += "\t\t<node id=\"{}\"/>\n".format(next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has-a</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Weights</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:initializer = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.W_init)
        if self.W_cons:
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:regularizer = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.W_cons)
        if self.W_reg:
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:constraint = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.W_reg)
        if self.use_bias:
            node_counter += 1
            next_node_id = "n" + str(node_counter)
            graph_nodes += "\t\t<node id=\"{}\"/>\n".format(next_node_id)
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has-a</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, next_node_id)
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Bias</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id)
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:initializer = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.b_init)
            if self.b_cons:
                graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:regularizer = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                    next_node_id, next_node_id, self.b_cons)
            if self.b_reg:
                graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:constraint = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                    next_node_id, next_node_id, self.b_reg)
        return node_counter, graph_nodes, graph_edges

class Dense(Learning_Layer):
    params = ['units', 'activation', 'use_bias', 'kernel_initializer', 'bias_initializer', 'kernel_regularizer', 'bias_regularizer','inputs']
    
    def __init__(self, name, input_layer_name, params):
        super(Dense, self).__init__(name, input_layer_name)
        self.units = params['units'] if 'units' in params else None
        self.activation = params['activation'] if 'activation' in params else None
        self.use_bias = params['use_bias'] if 'use_bias' in params else True
        self.W_init = params['kernel_initializer'] if 'kernel_initializer' in params else 'glorot_uniform'
        self.b_init = params['bias_initializer'] if 'bias_initializer' in params else 'zeros'
        self.W_cons = params['kernel_constraint'] if 'kernel_constraint' in params else None
        self.b_cons = params['bias_constraint'] if 'bias_constraint' in params else None
        self.W_reg = params['kernel_regularizer'] if 'kernel_regularizer' in params else None
        self.b_reg = params['bias_regularizer'] if 'bias_regularizer' in params else None
        self.clean_strings_names()

    def clean_strings_names(self):
        if self.activation:
            self.activation = self.activation.split('.')[-1]
        if self.W_init:
            self.W_init = self.W_init.split('.')[-1]
            if self.W_init in ['zeros', 'ones']:
                self.W_init = 'constant'
        if self.b_init:
            self.b_init = self.b_init.split('.')[-1]
        if self.W_cons:
            self.W_cons = self.W_cons.split('.')[-1]
        if self.b_cons:
            self.b_cons = self.b_cons.split('.')[-1]
        if self.W_reg:
            self.W_reg = self.W_reg.split('.')[-1]
        if self.b_reg:
            self.b_reg = self.b_reg.split('.')[-1]
            
    def set_input_size(self, input_size):
        self.input_size = input_size
        self.output_size = self.infer_output_size()

    def infer_output_size(self):
        batch_size = self.input_size[0]
        return [batch_size, self.units]
    
    def gen_graph_nodes_and_edges(self, node_id, node_counter):
        graph_nodes = ""
        graph_edges = ""
        #Layer
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:in_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.input_size))
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.output_size))
        for i in range(len(self.input_size)):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:in_dim{} = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, i, self.input_size[i])
        for i in range(len(self.output_size)):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dim{} = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, i, self.output_size[i])
        node_counter += 1
        next_node_id = "n" + str(node_counter)
        graph_nodes += "\t\t<node id=\"{}\"/>\n".format(next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has-a</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Parameters</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:type = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, 'dense')
        #additional Attributes
        node_counter += 1
        next_node_id = "n" + str(node_counter)
        graph_nodes += "\t\t<node id=\"{}\"/>\n".format(next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has-a</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Weights</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:initializer = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.W_init)
        if self.W_cons:
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:regularizer = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.W_cons)
        if self.W_reg:
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:constraint = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.W_reg)
        if self.use_bias:
            node_counter += 1
            next_node_id = "n" + str(node_counter)
            graph_nodes += "\t\t<node id=\"{}\"/>\n".format(next_node_id)
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has-a</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, next_node_id)
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Bias</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id)
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:initializer = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.b_init)
            if self.b_cons:
                graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:regularizer = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                    next_node_id, next_node_id, self.b_cons)
            if self.b_reg:
                graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:constraint = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                    next_node_id, next_node_id, self.b_reg)
        return node_counter, graph_nodes, graph_edges

class Pooling(Processing_Layer):
    params = ['pool_size', 'strides', 'padding', 'inputs']
    def __init__(self, name, input_layer_name, params):
        super(Pooling, self).__init__(name, input_layer_name)
        self.set_pool_size(params)
        self.set_strides(params)
        self.padding = params['padding'] if 'padding' in params else 'valid'

    def set_pool_size(self, params):
        pool_size = params['pool_size'] if 'pool_size' in params else [2, 2]
        if isinstance(pool_size, list):
            self.pool_size = pool_size
        elif isinstance(pool_size, tuple):
            self.pool_size = [pool_size[0], pool_size[1]]    
        else:
            self.pool_size = [pool_size, pool_size]

    def set_strides(self, params):
        strides = params['strides'] if 'strides' in params else None
        if strides is None or strides == 'None':
            strides = self.pool_size
        if isinstance(strides, list):
            self.strides = strides
        elif isinstance(strides, tuple):
            self.strides = [strides[0], strides[1]]    
        else:
            self.strides = [strides, strides]

    def set_input_size(self, input_size):
        self.input_size = input_size
        self.output_size = self.infer_output_size()

    def infer_output_size(self):
        if len(self.input_size) == 4:
            batch_size, in_height, in_width, in_depth = self.input_size
            if self.padding == 'valid':
                out_height = (in_height - self.pool_size[0]) / self.strides[0] + 1
                out_width = (in_width - self.pool_size[1]) / self.strides[1] + 1
            elif self.padding == 'same':
                out_height = (in_height - 1) / self.strides[0] + 1
                out_width = (in_width - 1) / self.strides[1] + 1
            elif self.padding == 'full':
                out_height = (in_height + self.pool_size[0] - 2) / self.strides[0] + 1
                out_width = (in_width + self.pool_size[1] - 2) / self.strides[1] + 1
            return [batch_size, int(out_height), int(out_width), in_depth]
        elif len(self.input_size) == 3:
            batch_size, in_steps, in_depth = self.input_size
            if self.padding == 'valid':
                out_steps = (in_steps - self.pool_size[0]) / self.strides[0] + 1
            elif self.padding == 'same':
                out_steps = (in_steps - 1) / self.strides[0] + 1
            elif self.padding == 'full':
                out_steps = (in_steps + self.pool_size[0] - 2) / self.strides[0] + 1
            return [batch_size, int(out_steps), in_depth]
    
    def gen_graph_nodes_and_edges(self, node_id, node_counter):
        graph_nodes = ""
        graph_edges = ""
        #Layer
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:in_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.input_size))
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.output_size))
        for i in range(len(self.input_size)):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:in_dim{} = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, i, self.input_size[i])
        for i in range(len(self.output_size)):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dim{} = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, i, self.output_size[i])
        node_counter += 1
        next_node_id = "n" + str(node_counter)
        graph_nodes += "\t\t<node id=\"{}\"/>\n".format(next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has-a</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Parameters</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:type = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.name.split('_')[0])
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:pool_dim1 = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.pool_size[0])
        if '2d' in self.name or '3d' in self.name:
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:pool_dim2 = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                    next_node_id, next_node_id, self.pool_size[1])
        if '3d' in self.name:
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:pool_dim3 = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                    next_node_id, next_node_id, self.pool_size[2])
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:strides_dim1 = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.strides[0])
        if '2d' in self.name or '3d' in self.name:
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:strides_dim2 = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.strides[1])
        if '3d' in self.name:
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:strides_dim3 = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.strides[2])
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:padding = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.padding)
        return node_counter, graph_nodes, graph_edges

class Dropout(Processing_Layer):
    params = ['rate', 'inputs']
    
    def __init__(self, name, input_layer_name, rate=0.5):
        super(Dropout, self).__init__(name, input_layer_name)
        self.rate = rate
        
    def set_input_size(self, input_size):
        self.input_size = input_size
        self.output_size = self.infer_output_size()

    def infer_output_size(self):
        return self.input_size
    
    def gen_graph_nodes_and_edges(self, node_id, node_counter):
        graph_nodes = ""
        graph_edges = ""
        #Layer
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:in_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.input_size))
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.output_size))
        for i in range(len(self.input_size)):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:in_dim{} = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, i, self.input_size[i])
        for i in range(len(self.output_size)):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dim{} = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, i, self.output_size[i])
        node_counter += 1
        next_node_id = "n" + str(node_counter)
        graph_nodes += "\t\t<node id=\"{}\"/>\n".format(next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has-a</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Parameters</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:type = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, 'dropout')
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:dropoutRate = real:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.rate)
        return node_counter, graph_nodes, graph_edges

class Batchnorm(Processing_Layer):
    params = ['inputs']
    
    def __init__(self, name, input_layer_name):
        super(Batchnorm, self).__init__(name, input_layer_name)
        
    def set_input_size(self, input_size):
        self.input_size = input_size
        self.output_size = self.infer_output_size()

    def infer_output_size(self):
        return self.input_size
    
    def gen_graph_nodes_and_edges(self, node_id, node_counter):
        graph_nodes = ""
        graph_edges = ""
        #Layer
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:in_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.input_size))
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.output_size))
        for i in range(len(self.input_size)):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:in_dim{} = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, i, self.input_size[i])
        for i in range(len(self.output_size)):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dim{} = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, i, self.output_size[i])
        node_counter += 1
        next_node_id = "n" + str(node_counter)
        graph_nodes += "\t\t<node id=\"{}\"/>\n".format(next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has-a</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Parameters</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:type = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, 'batch_norm')
        return node_counter, graph_nodes, graph_edges

class Reshape(Processing_Layer):
    params = ['target_shape', 'inputs']
    def __init__(self, name, input_layer_name, output_shape):
        super(Reshape, self).__init__(name, input_layer_name)
        self.output_shape = output_shape
        
    def set_input_size(self, input_size):
        self.input_size = input_size
        self.output_size = self.infer_output_size()

    def infer_output_size(self):
        output_shape = self.output_shape
        if -1 in output_shape:
            full_dims = np.prod(np.array(self.input_size))
            idx = output_shape.index(-1)
            output_shape[idx] = int(full_dims/np.prod(np.absolute(np.array(output_shape))))
            if np.prod(np.array(output_shape)) != full_dims:
                raise ValueError('The output shapes are not valid')
        return output_shape
    
    def gen_graph_nodes_and_edges(self, node_id, node_counter):
        graph_nodes = ""
        graph_edges = ""
        #Layer
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:in_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.input_size))
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.output_size))
        for i in range(len(self.input_size)):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:in_dim{} = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, i, self.input_size[i])
        for i in range(len(self.output_size)):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dim{} = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, i, self.output_size[i])
        node_counter += 1
        next_node_id = "n" + str(node_counter)
        graph_nodes += "\t\t<node id=\"{}\"/>\n".format(next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has-a</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Parameters</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:type = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, 'reshape')
        return node_counter, graph_nodes, graph_edges
    
class Flatten(Processing_Layer):
    params = ['inputs']
    def __init__(self, name, input_layer_name):
        super(Flatten, self).__init__(name, input_layer_name)
        
    def set_input_size(self, input_size):
        self.input_size = input_size
        self.output_size = self.infer_output_size()

    def infer_output_size(self):
        batch_size = self.input_size[0]
        return [batch_size, int(np.prod(self.input_size[1:]))]

    def gen_graph_nodes_and_edges(self, node_id, node_counter):
        graph_nodes = ""
        graph_edges = ""
        #Layer
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:in_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.input_size))
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.output_size))
        for i in range(len(self.input_size)):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:in_dim{} = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, i, self.input_size[i])
        for i in range(len(self.output_size)):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dim{} = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, i, self.output_size[i])
        node_counter += 1
        next_node_id = "n" + str(node_counter)
        graph_nodes += "\t\t<node id=\"{}\"/>\n".format(next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has-a</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Parameters</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:type = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, 'flatten')
        return node_counter, graph_nodes, graph_edges

class Activation(Processing_Layer):
    activations = ['elu', 'relu', 'selu', 'relu6', 'leakyrelu', 'tanh', 'sigmoid', 'softmax', 'linear']
    params  = ['inputs']

    def __init__(self, name, input_layer_name, activation):
        super(Activation, self).__init__(name, input_layer_name)
        self.activation = activation if activation else 'linear'
        self.non_linear = self.check_non_linearity()

    def check_non_linearity(self):
        if self.activation in ['identity']:
            return False
        return True

    def set_input_size(self, input_size):
        self.input_size = input_size
        self.output_size = self.infer_output_size()

    def infer_output_size(self):
        return self.input_size
    
    def gen_graph_nodes_and_edges(self, node_id, node_counter):
        graph_nodes = ""
        graph_edges = ""
        #Layer
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:in_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.input_size))
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dims = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
            node_id, node_id, len(self.output_size))
        for i in range(len(self.input_size)):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:in_dim{} = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, i, self.input_size[i])
        for i in range(len(self.output_size)):
            graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:out_dim{} = int:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, node_id, i, self.output_size[i])
        node_counter += 1
        next_node_id = "n" + str(node_counter)
        graph_nodes += "\t\t<node id=\"{}\"/>\n".format(next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>has-a</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>type:Parameters</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:type = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, 'activator')
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:activation = string:\"{}\"</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, self.activation)
        graph_edges += "\t\t<edge from=\"{}\" to=\"{}\">\n\t\t\t<attr name=\"label\">\n\t\t\t\t<string>let:nonLinear = bool:{}</string>\n\t\t\t</attr>\n\t\t</edge>\n".format(
                next_node_id, next_node_id, str(self.non_linear).lower())
        return node_counter, graph_nodes, graph_edges


import ast
from pprint import pprint
from ast import parse, NodeVisitor, literal_eval
from nodes_Keras import *

def main(program_name, input_size=[32, 28, 28, 1], output_size=[32, 10]):
    with open(f"{program_name}.py", "r") as source:
        source_code = [line for line in source.readlines()]
    with open(f"{program_name}.py", "r") as source:
        tree = ast.parse(source.read())
    dnn = DNN(input_size, output_size)
    ffc = Parse_Keras_Program(dnn)
    ffc.visit(tree)
    program = DNN_Program(ffc.DNN, ffc.learner)
    program.gen_graph_nodes_and_edges()
    program.write_graph_node(program_name)
            
def parse_arg(arg):
    if isinstance(arg, ast.Str):
        return arg.s
    elif isinstance(arg, ast.Name):
        value = None if arg.id == "None" else arg.id
        return value
    elif isinstance(arg, ast.NameConstant):
        return arg.value
    elif isinstance(arg, ast.Num):
        return arg.n
    elif isinstance(arg, ast.List):
        arg_list = []
        for elt in arg.elts:
            arg_list.append(parse_arg(elt))
        return arg_list
    elif isinstance(arg, ast.Tuple):
        arg_list = []
        for elt in arg.elts:
            arg_list.append(parse_arg(elt))
        return tuple(arg_list)
    elif isinstance(arg, ast.Attribute):
        value = parse_arg(arg.value)
        return value + "." + arg.attr
        #return (str(arg.value.attr) + "." + str(arg.attr))
    elif isinstance(arg, ast.Call):
        return [parse_call(arg.func), parse_params(arg.args, arg.keywords)]
    elif isinstance(arg, ast.UnaryOp):
        if isinstance(arg.op, ast.USub):
            return -1 * arg.operand.n
    else:
        return arg 

def parse_call(call):
    if isinstance(call, ast.Name):
        value = None if call.id == "None" else call.id
        return value
    elif isinstance(call, ast.Attribute):
        value = parse_call(call.value)
        return value + "." + call.attr
    elif isinstance(call, ast.Str):
        return call.s
    else:
        return call 

def parse_target(arg):
    if isinstance(arg, ast.Name):
        value = None if arg.id == "None" else arg.id
        return value
    elif isinstance(arg, ast.Tuple):
        arg_list = []
        for elt in arg.elts:
            arg_list.append(parse_target(elt))
        return ','.join(arg_list)
    elif isinstance(arg, ast.Attribute):
        return (str(arg.value.id) + "." + str(arg.attr))
    else:
        return arg

def parse_params(args, keywords):
    params = {}
    for arg_idx in range(len(args)):
        arg = args[arg_idx]
        params[arg_idx] = parse_arg(arg)
    for keyword in keywords:
        params[keyword.arg] = parse_arg(keyword.value)
    return params

    def visit_Return(self, node):
        if self.func_name != None:
            self.return_ref = node.value

class Parse_Keras_Program(NodeVisitor):

    def __init__(self, dnn):
        self.DNN = dnn
        self.learner = Learner()
        self.batch_size = self.DNN.input_size[0]
        self.learner.set_batch_size(self.batch_size)
        self.model = 'model'
        self.variables = {}
        self.input_layer_name = 'data'
        self.layers_count = 1

    def visit_Assign(self, node):
        asigned_str = parse_arg(node.value)
        asigned_str = asigned_str[0] if isinstance(asigned_str, list) and len(asigned_str) > 0 else asigned_str
        call_target = parse_target(node.targets[0])
        if asigned_str == 'Sequential':
            self.model = call_target
        else:
            self.variables[call_target] = asigned_str   
    
    def visit_Call(self, node):
        call_str = parse_call(node.func)
        call_params = parse_params(node.args, node.keywords)
        if call_str == self.model+'.add':
            layer_type = call_params[0][0].split('.')[-1]
            layer_params = call_params[0][1]
            self.change_params_values(layer_params)
            if is_DNN_layer(layer_type): 
                layer_params['inputs'] = self.input_layer_name
                layer_params = self.replace_args(layer_params)
                layer_name = layer_type.lower() + '_' + str(self.layers_count)
                self.DNN.append_layer(layer_name, layer_type, layer_params)
                self.input_layer_name = layer_name
                self.layers_count += 1       
        elif call_str == self.model+'.compile':
            self.DNN.compile()
            self.change_params_values(call_params)
            loss = Loss(call_params['loss'], 'output')
            loss.set_predictions_layer(self.DNN.get_last_layer_idx())
            self.learner.set_loss(loss)
            opt = Optimizer(call_params['optimizer'])
            self.learner.set_optimizer(opt)
            if 'metrics' in call_params and len(call_params['metrics'])>0:
                metric = call_params['metrics'][0]#Metric(name=call_params['metrics'][0])
                self.learner.set_metric(metric)
        elif call_str == self.model+'.fit':
            self.change_params_values(call_params)
            epochs = call_params['epochs'] if 'epochs' in call_params else 1
            batch_size = call_params['batch_size'] if 'batch_size' in call_params else self.batch_size
            self.learner.set_vars_init_location(node.func.lineno - 1)
            self.learner.set_train_loop(Train_Loop(node.func.lineno, node.func.lineno))
            self.learner.set_batch_size(batch_size)
            self.learner.set_epochs(epochs)

    def change_params_values(self, params):
        for k, v in params.items():
            if isinstance(v, str) and v in self.variables:
                params[k] = self.variables[v]
    
    def replace_args(self, params):
        def _hashable(v):
            """Determine whether `v` can be hashed."""
            try:
                hash(v)
            except TypeError:
                return False
            return True
        for p_key, p_value in params.items():
            if _hashable(p_value):
                params[p_key] = p_value
            else:
                params[p_key] = None
        return params

if __name__ == "__main__":
    program_name = 'Keras_examples'+os.path.sep+'clean_example'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[128, 20], output_size=[128, 10])
    program_name = 'Keras_examples'+os.path.sep+'deep_CNN'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[128, 224, 224, 3], output_size=[128, 2])
    program_name = 'Keras_examples'+os.path.sep+'deep_CNN_too_much_pooling'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[128, 4098, 4098, 3], output_size=[128, 2])
    program_name = 'Keras_examples'+os.path.sep+'deep_CNN_asymetry_blocks'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[128, 224, 224, 3], output_size=[128, 2])
    program_name = 'Keras_examples'+os.path.sep+'deep_CNN_ineffective_poolings'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[128, 224, 224, 3], output_size=[128, 2])
    program_name = 'Keras_examples'+os.path.sep+'SO_33969059'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[32, 3], output_size=[32, 1])
    program_name = 'Keras_examples'+os.path.sep+'SO_34311586'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[32, 2], output_size=[32, 1])
    program_name = 'Keras_examples'+os.path.sep+'SO_38584268'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[32, 32, 32, 3], output_size=[32, 10])
    program_name = 'Keras_examples'+os.path.sep+'SO_44322611'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[10, 224, 224, 3], output_size=[10, 2])
    program_name = 'Keras_examples'+os.path.sep+'SO_45378493'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[128, 10, 4], output_size=[128, 1])
    program_name = 'Keras_examples'+os.path.sep+'SO_50079585_1'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[16, 150, 150, 3], output_size=[16, 3])
    program_name = 'Keras_examples'+os.path.sep+'SO_50079585_2'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[16, 150, 150, 3], output_size=[16, 3])
    program_name = 'Keras_examples'+os.path.sep+'SO_53119432'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[24, 30, 1], output_size=[24, 1])
    program_name = 'Keras_examples'+os.path.sep+'SO_55776436'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[32, 18000, 1], output_size=[32, 6])
    program_name = 'Keras_examples'+os.path.sep+'SO_56103207'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[100, 28, 28, 1], output_size=[100, 4])
    program_name = 'Keras_examples'+os.path.sep+'SO_60566498'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[32, 224, 224, 3], output_size=[32, 3])
    program_name = 'Keras_examples'+os.path.sep+'SO_45120429'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[32, 32, 1], output_size=[32, 1])
    program_name = 'Keras_examples'+os.path.sep+'SO_58844149'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[32, 300, 300, 1], output_size=[32, 1])
    program_name = 'Keras_examples'+os.path.sep+'SO_51749207'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[8, 600, 1], output_size=[8, 1])
    program_name = 'Keras_examples'+os.path.sep+'SO_55731589'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[128, 32, 32, 3], output_size=[128, 1])
    program_name = 'Keras_examples'+os.path.sep+'SO_45711636'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[100, 1, 20, 56], output_size=[100, 10])
    program_name = 'Keras_examples'+os.path.sep+'SO_49117607'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[32, 28, 28, 1], output_size=[32, 10])
    program_name = 'Keras_examples'+os.path.sep+'SO_44184091'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[32, 500, 32], output_size=[32, 1])
    program_name = 'Keras_examples'+os.path.sep+'SO_61030068'
    print(f'----------------{program_name}-------------------')
    main(program_name, input_size=[32, 72, 1], output_size=[32, 28])
    


import ast
from pprint import pprint
from ast import parse, NodeVisitor, literal_eval
from nodes_TF import *

def main(program_name, input_size=[32, 28, 28, 1], output_size=[32, 10]):
    with open(f"{program_name}.py", "r") as source:
        source_code = [line for line in source.readlines()]
    with open(f"{program_name}.py", "r") as source:
        tree = ast.parse(source.read())
    ffc = Find_Function_Calls()
    ffc.visit(tree)
    model = DNN(input_size, output_size)
    train_op_ref, opt, loss = parse_assigns_with_call(model, ffc.assigns_with_call, tree)
    model.compile()
    loss.set_predictions_layer(model.get_last_layer_idx(loss.predictions))
    learner = Learner()
    learner.set_loss(loss)
    learner.set_optimizer(opt)
    learner.set_batch_size(input_size[0])
    fc = Find_Loops_And_Runs(source_code, train_op_ref)
    fc.visit(tree)
    learner.set_vars_init_location(fc.init_line)
    if len(fc.loops) > 0:
        main_loop = parse_main_loop(fc.loops, fc.train_line)
        train_loop = Train_Loop(main_loop['start'], main_loop['end'])
    else:
        train_loop = Train_Loop(0, 0)
    learner.set_train_loop(train_loop)
    program = DNN_Program(model, learner)
    program.gen_graph_nodes_and_edges()
    program.write_graph_node(program_name)

def parse_assigns_with_call(model, assigns_with_call, ast):
    opt_ref, loss_ref = '', ''
    for call_target, (call_str, call_params) in assigns_with_call.items():
        if is_optimizer(call_str.split('.')[-1]):
            opt_ref = call_target
            opt = Optimizer(name=call_str.split('.')[-1])
        if call_str == f'{opt_ref}.minimize':
            loss_ref = list(call_params.values())[0]
            labels, logits = list(assigns_with_call[loss_ref][1].values())
            loss_func = Loss(assigns_with_call[loss_ref][0].split('.')[-1], logits)
            train_op_ref = call_target
            while logits in assigns_with_call:
                func_name, params = assigns_with_call[logits]
                layer_type = func_name.split('.')[-1]
                if is_DNN_layer(layer_type):
                    model.prepend_layer(logits, layer_type, params)
                    logits = model.layers[0].input_layer_name
                else:
                    args = list(params.values())
                    fl = Find_Layers(func_name, args)
                    fl.visit(ast)
                    for layer_name, layer_type, params in reversed(fl.layers_data):
                        #print(layer_name, layer_type, params)
                        model.prepend_layer(layer_name, layer_type, params)
                    model.update_layers_name(fl.return_ref.id, logits)
                    logits = model.layers[0].input_layer_name
    return train_op_ref, opt, loss_func

def parse_main_loop(loops, train_line):
    loops = sorted(loops, key=lambda kv: kv['start']+kv['end'], reverse=True)
    for loop in loops:
        if train_line in range(loop['start'],loop['end']):
            return loop
            
def parse_arg(arg):
    if isinstance(arg, ast.Str):
        return arg.s
    elif isinstance(arg, ast.Name):
        value = None if arg.id == "None" else arg.id
        return value
    elif isinstance(arg, ast.NameConstant):
        return arg.value
    elif isinstance(arg, ast.Num):
        return arg.n
    elif isinstance(arg, ast.List):
        arg_list = []
        for elt in arg.elts:
            arg_list.append(parse_arg(elt))
        return arg_list
    elif isinstance(arg, ast.Tuple):
        arg_list = []
        for elt in arg.elts:
            arg_list.append(parse_arg(elt))
        return tuple(arg_list)
    elif isinstance(arg, ast.Attribute):
        value = parse_arg(arg.value)
        return value + "." + arg.attr
        #return (str(arg.value.attr) + "." + str(arg.attr))
    elif isinstance(arg, ast.Call):
        return parse_call(arg.func)
    elif isinstance(arg, ast.UnaryOp):
        if isinstance(arg.op, ast.USub):
            return -1 * arg.operand.n
    else:
        return arg 

def parse_call(call):
    if isinstance(call, ast.Name):
        value = None if call.id == "None" else call.id
        return value
    elif isinstance(call, ast.Attribute):
        value = parse_call(call.value)
        if isinstance(value, ast.BinOp):
            print(value.left.id, value.op, value.right.n)
        return value + "." + call.attr
    else:
        return call 

def parse_target(arg):
    if isinstance(arg, ast.Name):
        value = None if arg.id == "None" else arg.id
        return value
    elif isinstance(arg, ast.Tuple):
        arg_list = []
        for elt in arg.elts:
            arg_list.append(parse_target(elt))
        return ','.join(arg_list)
    elif isinstance(arg, ast.Attribute):
        return (str(arg.value.id) + "." + str(arg.attr))
    else:
        return arg

def parse_params(args, keywords):
    params = {}
    for arg_idx in range(len(args)):
        arg = args[arg_idx]
        params[arg_idx] = parse_arg(arg)
    for keyword in keywords:
        params[keyword.arg] = parse_arg(keyword.value)
    return params

def compute_interval(node):
    min_lineno = node.lineno
    max_lineno = node.lineno
    for node in ast.walk(node):
        if hasattr(node, "lineno"):
            min_lineno = min(min_lineno, node.lineno)
            max_lineno = max(max_lineno, node.lineno)
    return (min_lineno, max_lineno + 1)

class Find_Layers(NodeVisitor):

    def __init__(self, func_name, args):
        self.func_name = func_name
        self.return_ref = None
        self.args = args
        self.args_matches = {}
        self.layers_data = []
    
    def visit_FunctionDef(self, node):
        if self.func_name == None:
            self.generic_visit(node)
        else:
            if node.name == self.func_name:
                for i in range(len(node.args.args)):
                    param = node.args.args[i].arg
                    self.args_matches[parse_arg(param)] = self.args[i]
                # visit the children
                self.generic_visit(node)

    def replace_args(self, params):
        def _hashable(v):
            """Determine whether `v` can be hashed."""
            try:
                hash(v)
            except TypeError:
                return False
            return True
        for p_key, p_value in params.items():
            if _hashable(p_value) and p_value in self.args_matches:
                params[p_key] = self.args_matches[p_value]
        return params

    def visit_Assign(self, node):
        if isinstance(node.value, ast.Call):
            layer_type = (node.value.func.attr).lower()
            layer_name = parse_target(node.targets[0]).lower()
            if is_DNN_layer(layer_type): 
                args = node.value.args
                kwargs = node.value.keywords
                params = parse_params(args, kwargs)
                params = self.replace_args(params)
                self.layers_data.append((layer_name, layer_type, params))

    def visit_Return(self, node):
        if self.func_name != None:
            self.return_ref = node.value

class Find_Function_Calls(NodeVisitor):

    def __init__(self):
        self.assigns_with_call = {}

    def visit_Assign(self, node):
        if isinstance(node.value, ast.Call):
            call_str = parse_call(node.value.func)
            call_target = parse_target(node.targets[0])
            call_params = parse_params(node.value.args, node.value.keywords)
            self.assigns_with_call[call_target] = (call_str, call_params)

class Find_Loops_And_Runs(NodeVisitor):

    def __init__(self, source_code, train_op):
        self.source_code = source_code
        self.train_op = train_op
        self.sess_ref = ''
        self.train_line = None
        self.init_line = None
        self.loops = []

    def visit_Call(self, node):
        call_str = parse_call(node.func)
        if len(call_str.split('.'))==2 and 'Session'==call_str.split('.')[1]:
            if 'as' in self.source_code[node.func.lineno-1]:
                self.sess_ref = self.source_code[node.func.lineno-1].split('as')[1].replace(':','').strip()
            elif '=' in self.source_code[node.func.lineno-1]:
                self.sess_ref = self.source_code[node.func.lineno-1].split('=')[0].strip()
        run_call = self.sess_ref + '.run'
        if call_str == run_call:
            vars = parse_arg(node.args[0])
            if isinstance(vars, list) and 'tf.global_variables_initializer' in vars:
                self.init_line = node.func.lineno
            elif vars == 'tf.global_variables_initializer':
                self.init_line = node.func.lineno
            if isinstance(vars, list) and self.train_op in vars:
                self.train_line = node.func.lineno
            elif self.train_op == vars:
                self.train_line = node.func.lineno
    
    def visit_For(self, node):
        start, end = compute_interval(node)
        self.loops.append({'start':start, 'end':end})
        # visit the children
        self.generic_visit(node)
    
    def visit_While(self, node):
        start, end = compute_interval(node)
        self.loops.append({'start':start, 'end':end})
        # visit the children
        self.generic_visit(node)

if __name__ == "__main__":
    program_name = 'TF_lenet_disconnected_pooling_layers'
    main(program_name, input_size=[32, 28, 28, 1], output_size=[32, 10])


import tensorflow as tf
import click
import numpy as np
from tensorflow.keras.datasets import mnist

class Dataset:

    def __init__(self, X, y, batch_size):
        dset = tf.data.Dataset.from_tensor_slices((X, y))
        dset = dset.map(self.preprocess_example)
        dset = dset.shuffle(10000)
        dset = dset.batch(batch_size)
        self.dset = dset 
        self.iterator = self.dset.make_initializable_iterator()
        self.next_batch = self.iterator.get_next()

    def init(self, sess):
        sess.run(self.iterator.initializer)

    def preprocess_example(self, *example):
        new_image = tf.image.convert_image_dtype(tf.reshape(example[0], [28, 28, 1]), tf.float32)
        new_label = tf.cast(tf.one_hot(example[1], 10), tf.float32)
        return new_image, new_label

def build_model(input, scope='lenet', reuse=False, training=True):
    with tf.variable_scope(scope, reuse=reuse):
        conv1 = tf.layers.conv2d(inputs=input,
                                    filters=32,
                                    kernel_size=[15, 15],
                                    padding='valid',
                                    activation=tf.nn.relu)
        reg_conv1 = tf.layers.dropout(inputs=conv1, rate=0.35, training=training)
        conv1_norm = tf.layers.batch_normalization(reg_conv1, training=training)
        pool1 = tf.layers.max_pooling2d(inputs=conv1_norm, pool_size=[4, 4], strides=2)

        conv2 = tf.layers.conv2d(inputs=pool1, filters=64, kernel_size=[7, 7],
                                    padding='valid', activation=tf.nn.relu)
        reg_conv2 = tf.layers.dropout(inputs=conv2, rate=0.3, training=training)
        conv2_norm = tf.layers.batch_normalization(reg_conv2, training=training)
        pool2 = tf.layers.max_pooling2d(inputs=conv2_norm, pool_size=[4, 4], strides=2)

        pool2_flat = tf.layers.flatten(pool2)
        dense = tf.layers.dense(inputs=pool2_flat, units=512, activation=tf.nn.relu)
        dense_reg = tf.layers.dropout(inputs=dense, rate=0.25, training=training)
        dense_norm = tf.layers.batch_normalization(dense_reg, training=training)
        logits = tf.layers.dense(inputs=dense_norm, units=10, activation=None)
        return logits

def train(n_epochs, learning_rate, batch_size):
    X_train, y_train = mnist.load_data()[0]
    X_test, y_test = mnist.load_data()[1]
    train_dataset = Dataset(X_train, y_train, batch_size)
    test_dataset = Dataset(X_test, y_test, batch_size)
    X, y = train_dataset.next_batch
    y_pred = build_model(X, scope='lenet', reuse=False, training=True)
    train_loss = tf.losses.softmax_cross_entropy(y, y_pred)
    train_acc = tf.reduce_mean(tf.cast(tf.equal(tf.argmax(y, 1), tf.argmax(y_pred, 1)), tf.float32))
    update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
    with tf.control_dependencies(update_ops):
        optimizer = tf.train.AdamOptimizer(learning_rate)
        train_op = optimizer.minimize(train_loss)

    X_test, y_test = test_dataset.next_batch
    y_pred_test = build_model(X_test, reuse=True, training=False)
    test_loss = tf.losses.softmax_cross_entropy(y_test, y_pred_test)
    test_acc = tf.reduce_mean(tf.cast(tf.equal(tf.argmax(y_test, 1),tf.argmax(y_pred_test, 1)), tf.float32))

    with tf.Session() as sess:
        sess.run(tf.global_variables_initializer())
        for epoch in range(n_epochs):
            train_dataset.init(sess)
            epoch_errs = []
            epoch_accs = []
            try:
                while True:
                    _, err, acc = sess.run([train_op, train_loss, train_acc])
                    epoch_errs.append(err)
                    epoch_accs.append(acc)
            except tf.errors.OutOfRangeError:
                print(f"Epoch {epoch}:")
                print(f"  - Train err: {np.mean(epoch_errs)}")
                print(f"  - Train acc: {np.mean(epoch_accs)}")
                test_dataset.init(sess)
                test_errs = []
                test_accs = []
                try:
                    while True:
                        err, acc = sess.run([test_loss, test_acc])
                        test_errs.append(err)
                        test_accs.append(acc)
                except tf.errors.OutOfRangeError:
                    epoch_test_err, epoch_test_acc = np.mean(test_errs), np.mean(test_accs) 
                print(f"  - Test err: {epoch_test_err}")
                print(f"  - Test acc: {epoch_test_acc}")

@click.command()
@click.option('--n-epochs', type=int, default=5)
@click.option('--lr', type=float, default=3e-4)
@click.option('--batch-size', type=int, default=32)
def main(n_epochs, lr, batch_size):
    train(n_epochs, lr, batch_size)

if __name__ == '__main__':
    main()


import parser_TF
import parser_Keras
import os
import grooveParser
import sys
import ast

def parserType(fileName):
    try:
        source = open(f"{fileName}.py", "r")
    except:
        raise SystemExit("The file doesn't exist or it isn't a Python script ...")
    
    tree = ast.parse(source.read())
    tree_body = tree.body
    for item in tree_body:
        if isinstance(item, ast.Import):
            if item.names[0].name == "keras":
                return "keras"
            elif item.names[0].name == "tensorflow":
                return "tf"
        if isinstance(item, ast.ImportFrom):
            if "keras" in item.module:
                return "keras"
            elif "tensorflow" in item.module:
                return "tf"

def parsDnnScript(fileName, parser = "tf" , inputSize=[128, 20], outputSize=[128, 10], groovePath = "groove-5_7_4-bin/groove-5_7_4/bin/", grammarName="DNN-metamodel"):
    FileNameWithoutPath = fileName.split(os.path.sep)[-1]
    grooveOutputFileName = f"{FileNameWithoutPath}GrooveOut"
    parser_type = parser

    if parser_type == "keras":
        parser = "keras"

        try:
            print("Generating model (graph)  ... \n")
            parser_Keras.main(fileName, input_size=inputSize, output_size=outputSize)
            print("Running Model Checker (Groove) ... \n")
            os.system(
                f'java -jar {groovePath}Generator.jar -f Keras_graphs/{grooveOutputFileName}.gst -s bfs {grammarName}.gps Keras_graphs/{FileNameWithoutPath}.gst')
        except:
            return f"{FileNameWithoutPath}.py\n\rError: input file is not valid or not match with selected parser type"


    elif parser_type == "tf":
        parser = "tf"

        try:
            print("Generating model (graph)  ... \n")
            parser_TF.main(fileName, input_size= inputSize, output_size=outputSize)
            print("Running Model Checker (Groove) ... \n")
            os.system(
                f'java -jar {groovePath}Generator.jar -f TF_graphs/{grooveOutputFileName}.gst -s bfs {grammarName}.gps TF_graphs/{FileNameWithoutPath}.gst')
        except:
            return f"{FileNameWithoutPath}.py\n\rError: input file is not valid or not match with selected parser type"


    try:
        result = f"{FileNameWithoutPath}.py\n\r" + grooveParser.main(grooveOutputFileName, parserType = parser)
        return result
    except:
        return f"{FileNameWithoutPath}.py\n\rError: input file is not valid or not match with selected parser type"

def checkInOutSize(inOut):

    if inOut[0] != '[' or inOut[len(inOut)-1] != ']':
        return False
    inOut = inOut[1:len(inOut)-1]
    inOutList = inOut.split(",")
    inOutList = list(map(int, inOutList))

    return inOutList

def main(file_folderName,inputSize,outputSize,parserName,resultFileName):
    if inputSize == False or outputSize==False:
        raise SystemExit("Error : Input and output size should be entered as array([x1, x2, x3, ...]).")
    parserName=parserName.lower()
    if parserName != "tf" and parserName != "keras":
        raise SystemExit("Error : Parser type should be 'tf' or 'keras'")
    try:
        outputFile = open(f"{resultFileName}.txt" ,"w", encoding="ISO-8859-1")
    except IOError:
        raise SystemExit("Error : output filename should meet host operating system filename rules")
    extention = os.path.splitext(file_folderName)[1]
    if extention == ".py":
        output = parsDnnScript(fileName = os.path.splitext(file_folderName)[0], inputSize = inputSize ,outputSize = outputSize , parser=parserName)
        outputFile.write(output)
    else:
        raise SystemExit("Error : input should be a python script file")
    
if __name__ == '__main__':
    main(file_folderName="mix_bugs_1.py",inputSize=[32,28,28,1],outputSize=[32,10],parserName="tf",resultFileName="result")

